/*

  Copyright 2000 Lunarmobiscuit, morgan

  file:			main.cpp

*/

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include "DlPortIO.h"

/*--- Constants ---*/

#define kLPT1			0x378

#define kPPData			kLPT1+0

#define kPPStatus		kLPT1+1
#define kPBSY			0x80

#define kPPControl		kLPT1+2
#define kPWE			0x01
#define kPDS			0x02
#define kPINT			0x04
#define kPSEL			0x08
#define kPREAD			0x20

#define true			(1 == 1)
#define false			(1 == 0)

// Command to execute
typedef enum {
	cmdError,
	cmdWriteByte,
	cmdWriteWord,
	cmdWriteLongs,
	cmdReadByte,
	cmdReadWord,
	cmdReadLongs,
	cmdProcess,
	cmdDebug,
	cmdTiming
} Command;

// Command line parameter values
typedef struct {
	unsigned long	long1;
	unsigned long	long2;
	unsigned short	word1;
	unsigned char	byte1;
} CommandValues;


/*--- Signals --- */

//	Signal				FPGA	Register.Bit	Comments

//	PD0					P4		378.0			Parallel Port Data
//	PD1					P5		378.1
//	PD2					P6		378.2
//	PD3					P7		378.3
//	PD4					P8		378.4
//	PD5					P16		378.5
//	PD6					P18		378.6
//	PD7					P17		378.7

//	PWE (Write Enable)	P?		37A.0			Writing Data
//	PDS (Data Strobe)	P14		37A.1			Data is Ready (Hold for 0.5us)


/*
	Initialize the Parallel Port registers
*/
void InitParallelPort(void)
{
	DlPortWritePortUchar(kPPControl, 0);
} // InitParallelPort


/*
	Stop using the Parallel Port registers
*/
void ExitParallelPort(void)
{
	DlPortWritePortUchar(kPPControl, 0);
} // ExitParallelPort


/*
	Debug the Parallel Port device (by strobing PSEL)
*/
void DebugParallelPort(void)
{
	unsigned char status;

	DlPortWritePortUchar(kPPControl, kPSEL|kPWE);
	DlPortWritePortUchar(kPPData, 0xDD);
	do { status = DlPortReadPortUchar(kPPStatus); }
	while ((status & kPBSY) == 0x00);
	Sleep(1*1000);

	DlPortWritePortUchar(kPPControl, kPSEL|kPWE|kPDS);
	do { status = DlPortReadPortUchar(kPPStatus); }
	while ((status & kPBSY) == kPBSY);
	Sleep(1*1000);

	DlPortWritePortUchar(kPPControl, kPSEL|kPWE);
	do { status = DlPortReadPortUchar(kPPStatus); }
	while ((status & kPBSY) == 0x00);
	Sleep(1*1000);
} // DebugParallelPort


/*
	Write a value out the Parallel Port
*/
void WriteParallelPort(unsigned char value)
{
	unsigned char status;

	// Write-enable the Parallel Port (wait for no ACK)
	DlPortWritePortUchar(kPPControl, kPWE);
	DlPortWritePortUchar(kPPData, value);
	do { status = DlPortReadPortUchar(kPPStatus); }
	while ((status & kPBSY) == 0x00);

	// Strobe the Data-ready line (wait for ACK)
	DlPortWritePortUchar(kPPControl, kPWE|kPDS);
	do { status = DlPortReadPortUchar(kPPStatus); }
	while ((status & kPBSY) == kPBSY);

	// Unset Data-ready line (wait for no ACK)
	DlPortWritePortUchar(kPPControl, kPWE);
	do { status = DlPortReadPortUchar(kPPStatus); }
	while ((status & kPBSY) == 0x00);
} // WriteParallelPort


/*
	Read a value from the Parallel Port
*/
unsigned char ReadParallelPort(void)
{
	unsigned char value;
	unsigned char status;

	// Read-enable the Parallel Port (wait for no DATA)
	DlPortWritePortUchar(kPPControl, kPREAD|kPINT);
	do { status = DlPortReadPortUchar(kPPStatus); }
	while ((status & kPBSY) == 0x00);

	// Strobe the Data-ready line (wait for DATA)
	DlPortWritePortUchar(kPPControl, kPREAD|kPINT|kPDS);
	do { status = DlPortReadPortUchar(kPPStatus); }
	while ((status & kPBSY) == kPBSY);
	value = DlPortReadPortUchar(kPPData);

	// Unset Data-ready line (wait for no DATA)
	DlPortWritePortUchar(kPPControl, kPREAD|kPINT);
	do { status = DlPortReadPortUchar(kPPStatus); }
	while ((status & kPBSY) == 0x00);

	return value;
} // ReadParallelPort


/*
	Process an Array of Data
*/
void ProcessData(void)
{
	// Test Input Data Array
	unsigned char arrLength = 11;
	unsigned char inpLength = arrLength + 1;
	const unsigned char arrInput[] = {arrLength, 0, 0x1, 0x2,0x4,0x8,0x10,0x20,0x40,0x80,0xAA,0x55};

	// Output Data Arrays
	unsigned char swOutput[sizeof(arrInput) / sizeof(char)], hwOutput[sizeof(arrInput) / sizeof(char)];

	unsigned char arrAverage, i;	
	int arrSum = 0;
	long swStart, swEnd, hwStart, hwEnd;

	//Start Processing
	printf("\nStart Processing...\n\n");

	// Print Result Table Header
	printf("Processed Array Data\n");
	printf("Index    Input    S/W Output    H/W Output\n");
	printf("___________________________________________\n");
	printf("\n");

	// Compute Software Results
	swStart = GetTickCount();

	for (i = 1; i < inpLength; i++)
	{
		arrSum = arrSum + arrInput[i];    // Sum Data
	}

	arrAverage = arrSum / arrLength;    // Calculate Average

	for (i = 0; i < inpLength; i++)
	{
		// Set Output arrLength Value
		if (i == 0)
		{
			swOutput[i] = arrInput[i];
		}
		// Normalize Data
		else if (arrInput[i] < arrAverage)
		{
			swOutput[i] = arrInput[i] + ((arrAverage - arrInput[i]) / 2);
		}
		else if (arrInput[i] > arrAverage)
		{
			swOutput[i] = arrInput[i] - ((arrInput[i] - arrAverage) / 2);
		}
		else
		{
			swOutput[i] = arrAverage;
		}
	}

	swEnd = GetTickCount();
	// End of Software Computation

	// Compute Hardware Results
	hwStart = GetTickCount();

	// Write Data to Hardware
	for (i = 0; i < inpLength; i++)
	{
		WriteParallelPort(arrInput[i]);
	}

	// Wait for Hardware to Begin Processing
	// Sleep(1*1000);

	// Read Processed Data from Hardware
	for (i = 0; i < inpLength; i++)
	{
		hwOutput[i] = ReadParallelPort();
	}

	hwEnd = GetTickCount();
	// End of Hardware Computation

	// Print Results
	for (i = 0; i < inpLength; i++)
	{
		printf("%02x", i);
		printf("       ");
		printf("%02x", arrInput[i]);
		printf("       ");
		printf("%02x", arrInput[i]);
		printf("            ");
		printf("%02x", hwOutput[i]);
		printf("\n");
	}

	// Report Processing Times
	printf("\nProcessing Time:");
	printf("  ");
	printf("%ldus", (swEnd - swStart)*1000);
	printf("           ");
	printf("%ldus\n", (hwEnd - hwStart)*1000);

	// End Processing
	printf("\nEnd Processing.\n\n");

} // ProcessData


/*
	Compute the time it takes to send one byte out the Parallel Port
*/
#define TIME_KBYTES		100
#define TIME_LOOPS		(1024*TIME_KBYTES)

void TimeParallelPort()
{
	int			start, end;
	int			i;

	start = GetTickCount();
	for (i=0; i<TIME_LOOPS; i++)
		WriteParallelPort((unsigned char)i);
	end = GetTickCount();

	printf("%ldms/%ldKB = %ld.%ldus/B = %ld.%ldKB/s\n",
		(end-start), TIME_KBYTES,
		(end-start)*1000/TIME_LOOPS, ((end-start)*10000/TIME_LOOPS)%10,
		TIME_KBYTES*1000/(end-start), (TIME_KBYTES*10000/(end-start))%10);
} // TimeParallelPort


/*
	Compute the value of a hex character
*/
int HexValue(char h)
{
	if ((h >= '0') && (h <= '9')) return h - '0';
	else if ((h >= 'A') && (h <= 'F')) return h - 'A' + 10;
	else if ((h >= 'a') && (h <= 'f')) return h - 'a' + 10;
	else return -1;
} // HexValue


/*
	Convert a hexidecimal string into an integer
*/
int hextoi(char* IN str)
{
	int				value = 0;
	int				hex;

	if (str == NULL) return 0;

	while (*str) {
		hex = HexValue(*str);
		if (hex < 0) return value;
		value = (value * 16) + hex;
		str += 1;
	}

	return value;
} // hextoi


/*
	Convert a hexidecimal string into an integer
*/
void Usage(char *str)
{	printf("Usage: PPortC <long1> <long2>\n");
	printf("Usage: PPortC w<word>\n");
	printf("Usage: PPortC +<byte>\n");
	printf("Usage: PPortC r{b,w,l}\n");
	printf("Usage: PPortC -process\n");
	printf("Usage: PPortC -debug\n");
	printf("Usage: PPortC -timing\n");
	printf("  <long> := 8-digit hex or d########## for decimal\n");
	printf("  <word> := 4-digit hex or d##### for decimal\n");
	printf("  <byte> := 2-digit hex or d### for decimal\n");
	if (str) printf("%s\n", str);
	exit(1);
} // Usage


/*
	Parse an argument
*/
Command ParseArgs(int argc, char* argv[], CommandValues* values)
{
	char			*arg = *argv;
	int				value;

	// Initiailze the values
	values->long1 = values->long2 = 0;
	values->word1 = 0;
	values->byte1 = 0;

	// Check for arguments
	if (argc == 0) Usage(NULL);

	if (*arg == 0) return cmdError;
	else if (*arg == '-') {
		if (argc > 1) return cmdError;
		if (strcmp(arg+1, "process") == 0) return cmdProcess;
		else if (strcmp(arg+1, "debug") == 0) return cmdDebug;
		else if (strcmp(arg+1, "timing") == 0) return cmdTiming;
		else return cmdError;
	}
	// Read values
	else if ((*arg == 'r') || (*arg == 'R')) {
		if (*(arg+1) == 'b') return cmdReadByte;
		else if (*(arg+1) == 'w') return cmdReadWord;
		else return cmdReadLongs;
	}
	// One byte
	else if (*arg == '+') {
		if (argc > 1) return cmdError;
		if (*(arg+1) == 'd') value = atoi(arg+2);
		else value = hextoi(arg+1);
		if (value > 255) Usage("Error: The value exceeds a byte");
		else values->byte1 = (unsigned char)value;
		return cmdWriteByte;
	}
	// One word
	else if (*arg == 'w') {
		if (argc > 1) return cmdError;
		if (*(arg+1) == 'd') value = atoi(arg+2);
		else value = hextoi(arg+1);
		if (value > 65535) Usage("Error: The value exceeds a word");
		else values->word1 = (unsigned short)value;
		return cmdWriteWord;
	}
	// Two longs
	else {
		if (argc != 2) return cmdError;

		arg = argv[0];
		if (*arg == 'd') values->long1 = atoi(arg+1);
		else values->long1 = (unsigned long)hextoi(arg);

		arg = argv[1];
		if (*arg == 'd') values->long2 = atoi(arg+1);
		else values->long2 = (unsigned long)hextoi(arg);
		return cmdWriteLongs;
	}

	return cmdError;
} // ParseArgs


/*
	Entry point
*/
int main(int argc, char* argv[])
{
	Command			cmd;
	CommandValues	values;
	
	cmd = ParseArgs(argc-1, argv+1, &values);
	if (cmd == cmdError) Usage(NULL);

	// Initialzie the Parallel Port
	InitParallelPort();

	// Execute the command
	switch (cmd) {
		case cmdWriteByte:
			WriteParallelPort(values.byte1);
			break;

		case cmdWriteWord:
			WriteParallelPort((unsigned char)(values.word1 >> 8));
			WriteParallelPort((unsigned char)(values.word1 >> 0));
			break;

		case cmdWriteLongs:
			WriteParallelPort((unsigned char)(values.long1 >> 24));
			WriteParallelPort((unsigned char)(values.long1 >> 16));
			WriteParallelPort((unsigned char)(values.long1 >> 8));
			WriteParallelPort((unsigned char)(values.long1 >> 0));

			WriteParallelPort((unsigned char)(values.long2 >> 24));
			WriteParallelPort((unsigned char)(values.long2 >> 16));
			WriteParallelPort((unsigned char)(values.long2 >> 8));
			WriteParallelPort((unsigned char)(values.long2 >> 0));
			break;

		case cmdReadByte:
			values.byte1 = ReadParallelPort();
			printf("< %02x\n", values.byte1);
			break;

		case cmdReadWord:
			values.word1 = ReadParallelPort() << 8;
			values.word1 |= ReadParallelPort();
			printf("< %04x\n", values.word1);
			break;

		case cmdReadLongs:
			values.long1 = ReadParallelPort() << 24;
			values.long1 |= ReadParallelPort() << 16;
			values.long1 |= ReadParallelPort() << 8;
			values.long1 |= ReadParallelPort();
			printf("< %08x\n", values.long1);

			values.long2 = ReadParallelPort() << 24;
			values.long2 |= ReadParallelPort() << 16;
			values.long2 |= ReadParallelPort() << 8;
			values.long2 |= ReadParallelPort();
			printf("< %08x\n", values.long2);
			break;

		case cmdProcess:
			ProcessData();
			break;

		case cmdDebug:
			DebugParallelPort();
			break;

		case cmdTiming:
			TimeParallelPort();
			break;
	}

	// Stop using the Parallel Port
	ExitParallelPort();
	return 0;
}
